/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    scoreDetections.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
** DESCRIPTION:
**   Given two XML object files, computes the precision and recall.
*****************************************************************************/

#include <cstdlib>
#include <string>
#include <limits>
#include <iomanip>
#include <fstream>
#include <map>
#include <vector>

#include "svlBase.h"
#include "svlVision.h"

using namespace std;

// type definitions and function prototypes ----------------------------------



// main ----------------------------------------------------------------------

void usage()
{
    cerr << SVL_USAGE_HEADER << endl;
    cerr << "USAGE: ./scoreDetections [OPTIONS] (<groundtruth xml> <results xml>)+" << endl;
    cerr << "OPTIONS:" << endl
	 << "  -area <num>       :: area overlap for positive hit (default 0.5)" << endl
	 << "  -threshold <num>  :: threshold for positive classification (default 0.5)" << endl
	 << "  -pr <filestem>    :: generate precision-recall curves" << endl
	 << "  -o <filestem>     :: output detections (true and false)" << endl
	 << "  -object <name>    :: specifies one object of interest" << endl
	 << "  -match            :: ignores frames that occur in one file but not the other" << endl
         << SVL_STANDARD_OPTIONS_USAGE
	 << endl;
}

int main(int argc, char *argv[])
{
    // read commandline parameters
    const int NUM_REQUIRED_PARAMETERS = 2;

    const char *gtFilename = NULL;
    const char *xmlFilename = NULL;

    const char *object = NULL;

    const char *prFilestem = NULL;
    const char *outputFilestem = NULL;
    double areaRatio = 0.5;
    double threshold = 0.5;
    bool onlyMatchingFrames = false;

    char **args = argv + 1;
    while (--argc > NUM_REQUIRED_PARAMETERS) {
        SVL_PROCESS_STANDARD_OPTIONS(args, argc)
	else if (!strcmp(*args, "-area")) {
	    areaRatio = atof(*(++args)); argc--;
	} else if (!strcmp(*args, "-threshold")) {
	    threshold = atof(*(++args)); argc--;
 	} else if (!strcmp(*args, "-pr")) {
	    prFilestem = *(++args); argc--;
        } else if (!strcmp(*args, "-prPoints")) {
            SVL_LOG(SVL_LOG_WARNING, "-prPoints flag has been deprecated");
            argc--;
	} else if (!strcmp(*args, "-o")) {
	    outputFilestem = *(++args); argc--;
	} else if (!strcmp(*args, "-object")) {
	    object = *(++args); argc--;
	} else if (!strcmp(*args, "-match")) {
            onlyMatchingFrames = true;
	} else if (**args == '-') {
            SVL_LOG(SVL_LOG_FATAL, "unrecognized option " << *args);
	} else {
	    break;
	}
	args++;
    }

    if (argc < NUM_REQUIRED_PARAMETERS) {
	usage();
	return -1;
    }

    // Accumulate counts per object class, e.g. objectScores["mug"]
    // holds the score of every mug detection and whether it is a
    // true positive (1) or not (0).

    svlObjectDetectionAnalyzer *d;
    if (object == NULL) {
      d = new svlMultObjectsDetectionAnalyzer();
    } else {
      d = new svlObjectDetectionAnalyzer(true, string(object));
    }

    for (unsigned i = 0; i < (unsigned)(argc / 2); i++) {
        svlObject2dSequence gtObjectList;
        svlObject2dSequence objectList;

	gtFilename = args[2 * i];
	xmlFilename = args[2 * i + 1];

	SVL_LOG(SVL_LOG_VERBOSE, "Reading from " << gtFilename << " and " << xmlFilename);
	
	if (!d->loadGt(gtFilename)) continue;
	if (!d->loadDetections(xmlFilename)) continue;

 	// output detections
	if (outputFilestem) {
	  string filename = string(outputFilestem) + string(".") + toString(i) + string(".xml");
	  d->outputDetections(filename.c_str(), threshold);
	}

        // accumulate counts
        d->accumulateScores(onlyMatchingFrames, areaRatio);
    }
   
    // show results
    d->printResults((float)threshold);

    // construct PR curve
    if (prFilestem != NULL) {
	SVL_LOG(SVL_LOG_VERBOSE, "Generating precision-recall curves...");
	d->createPRcurves();
	d->writePRcurves(prFilestem);
	SVL_LOG(SVL_LOG_VERBOSE, "...done");
    }

    delete d;

    return 0;
}


