/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlCudaML.cu
** AUTHOR(S):   Paul Baumstarck <pbaumstarck@stanford.edu>
** DESCRIPTION:
**  Cuda machine learning functions.
**
*****************************************************************************/

#include "svlCudaCommon.h"

// Evaluate a decision tree on some data.
/*__global__ void dtree_eval_slice(float *dtree, int n_trees, float *feats, int feature_stride, float *res)
{
__shared__ float data[ 512 ];

// Use threadIdx.x for tree index.
if ( threadIdx.x >= n_trees )
return;

// Read in tree constants.
const int tix = 11*threadIdx.x;
const int root_ix = ((int*)dtree)[tix],
l_ix    = ((int*)dtree)[tix + 2],
r_ix    = ((int*)dtree)[tix + 6];
const float root_les = dtree[tix + 1],
l_les    = dtree[tix + 3],
r_les    = dtree[tix + 7],
ll_val   = dtree[tix + 4],
lr_val   = dtree[tix + 5],
rl_val   = dtree[tix + 8],
rr_val   = dtree[tix + 9],
r_val    = dtree[tix + 10];

// Read in data.
if ( threadIdx.x < feature_stride ) {
data[threadIdx.x] = feats[threadIdx.x];
}

__syncthreads();

// Evaluate model.
float gf = 0.0f;
if ( data[root_ix] <= root_les ) {
// Check left child.
if ( data[l_ix] <= l_les )
gf = ll_val;
else
gf = lr_val;
} else if ( r_ix >= 0 ) {
// Check right child.
if ( data[r_ix] <= r_les )
gf = rl_val;
else
gf = rr_val;
} else {
// Just the right value.
gf = r_val;
}

res[threadIdx.x] = gf;
}
*/

__global__ void dtree_eval_slice(float *dtree, int n_trees, float *feats, int feature_len, int feature_stride, int n_fvs, int execs_per_block)
{
	// Max tree size
	__shared__ float data[ 512 ];

	// Use threadIdx.x for tree index.
	if ( threadIdx.x >= n_trees )
		return;

	// Read in tree constants.
	const int tix = 11*threadIdx.x;
	const int root_ix = ((int*)dtree)[tix],
			  l_ix    = ((int*)dtree)[tix + 2],
			  r_ix    = ((int*)dtree)[tix + 6];
	const float root_les = dtree[tix + 1],
				l_les    = dtree[tix + 3],
				r_les    = dtree[tix + 7],
				ll_val   = dtree[tix + 4],
				lr_val   = dtree[tix + 5],
				r_val    = dtree[tix + 8],
				rl_val   = dtree[tix + 9],
				rr_val   = dtree[tix + 10];

	// Process all of the results assigned to us.
	int feature_n = blockIdx.x;
	for ( int i=0; i<execs_per_block; ++i ) {
		// Check if we have run out of features.
		if ( feature_n >= n_fvs )
			break;

		// Read in data.
		const int feature_ix = IMUL(feature_stride, feature_n);
		if ( blockDim.x >= feature_len ) {
			// More than enough threads for every feature.
			if ( threadIdx.x < feature_len ) { // TOTRY paulb: Might be able to take out checks and see if they speed anything appreciably.
				data[threadIdx.x] = feats[ feature_ix + threadIdx.x ];
			}
		} else {
			int j = threadIdx.x;
			do {
				data[j] = feats[ feature_ix + j ];
				j += blockDim.x;
				// __syncthreads(); ? To make the threads update in lockstep?
			} while ( j < feature_len );
		}

		// Sync after all data reads.
		__syncthreads();

		// Evaluate model.
		float gf = 0.0f;
		if ( data[root_ix] <= root_les ) {
			// Check left child.
			if ( data[l_ix] <= l_les )
				gf = ll_val;
			else
				gf = lr_val;
		} else if ( r_ix >= 0 ) {
			// Check right child.
			if ( data[r_ix] <= r_les )
				gf = rl_val;
			else
				gf = rr_val;
		} else {
			// Just assign the right value.
			gf = r_val;
		}

		// Move feature n through grid size.
		feature_n += gridDim.x;

		// Sync after we have read in features.
		__syncthreads();

		// Overwrite features with components of results.
		feats[ feature_ix + threadIdx.x ] = gf;
	}
}

void svlCudaDTreeEvaluate(svlCudaLine &dtree, int n_trees, svlCudaLine &features, int feature_len, int feature_stride, int n_fvs, int execs_per_block)
{
	//cout << "Running tree evaluation.\n";
	dim3 block(1), grid(1);
	if ( n_trees > 512 ) {
		// Too many trees to put all in one block; use blocking in y.
		// ...
	} else {
		// One thread for every tree in the "decision forest."
		block.x = iAlignUp(n_trees, 16);
		// Each block will process execs_per_block results, so only grid in x for the remainder.
		grid.x = iDivUp( n_fvs, execs_per_block );
	}
	/*printf("block: %d,%d\n", block.x, block.y);
	printf("grid: %d,%d\n", grid.x, grid.y);
	printf("execs: %d; stride: %d; len: %d\n", execs_per_block, feature_stride, feature_len);*/
	dtree_eval_slice<<< grid, block >>>(dtree.data, n_trees, features.data, feature_len, feature_stride, n_fvs, execs_per_block);
}


//// Allocate space for features and results.
//int n_fvs = 21;
//
//int getFeatureStride(int feature_len) {
//	int feature_stride = -1;
//	if ( feature_len <= 16 ) {
//		feature_stride = 16;
//	} else if ( feature_len <= 128 )	
//		feature_stride = pow( 2.0, ceil( log( feature_len ) / log( 2.0 ) ) );
//	} else {
//		feature_stride = iAlignUp(feature_len, 128);
//	}
//	return feature_stride;
//}
//// WARNING: If n_fvs == feature_stride, then the scan algorithm will sum up one element too few ...
//

// Collect results from feature computation.
__global__ void collect_results_slice(float *features, float *results, int ix, int n_fvs, int feature_stride)
{
	const int x = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;
	if ( x >= n_fvs )
		return;

	results[x] = features[ IMUL(x, feature_stride) + ix ];
}

// Collect feature results into res (by reading to end of every n_trees'th entry).
void svlCudaDTreeCollectResults(svlCudaLine *features, svlCudaLine *res, int n_trees, int n_fvs, int feature_stride)
{
	dim3 block(256);
	dim3 grid( iDivUp(n_fvs, block.x) );
	if ( n_trees == feature_stride ) {
		// This means we did cumsum-style integration so as not to overrun feature_stride, so the reference index is one less.
		--n_trees;
	}
	//printf("Collecting n: %d; fstride: %d; fvs: %d\n", n_trees, feature_stride, n_fvs);
	collect_results_slice<<< grid, block >>>(features->data, res->data, n_trees, n_fvs, feature_stride);
}


