/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlCudaMath.cu
** AUTHOR(S):   Paul Baumstarck <pbaumstarck@stanford.edu>
** DESCRIPTION:
**  Cuda math functions.
**
*****************************************************************************/

#include "svlCudaCommon.h"
#include <limits>

inline bool sizesMatch(svlCudaCoveringGrid *cudagrid, dim3 &grid) {
	return cudagrid->grid_x == grid.x
		&& cudagrid->grid_y == grid.y;
}

// Device memory.
//__device__ __constant__ float covering_grid[ 1536 ]; // For holding bits on whether to operation on this block or not.
__device__ __constant__ float covering_grid2[ 1536 ]; // For holding bits on whether to operation on this block or not.

// Perform <a -= b;>.
__global__ void subtract_slice(float *a, int a_pitch, float *b, int b_pitch, int data_w, int data_h)
{
	const int x = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;	// x index to patch.
	const int y = IMUL(blockDim.y, blockIdx.y) + threadIdx.y;	// y index to patch.
	if ( x >= data_w || y >= data_h )
		return;
	
	float *a_out = PITCH_GET(a, a_pitch, y);
	const float *b_out = PITCH_GET(b, b_pitch, y);
	/*float gf = a_out[x];
	a_out[x] = gf - b_out[x];*/
	a_out[x] -= b_out[x];
}

// Perform <a -= b;>.
void svlCudaSubtract(svlCudaPitch &a, svlCudaPitch &b, int img_w, int img_h)
{
	if ( img_w == -1 ) img_w = a.w;
	if ( img_h == -1 ) img_h = a.h;
	dim3 block(16, 16);
	dim3 grid( iDivUp(img_w, block.x), iDivUp(img_h, block.y) );
	subtract_slice<<< grid, block >>>(a.data, a.pitch, b.data, b.pitch, img_w, img_h);
}

// Decimate in x and y.
__global__ void decimate_slice(float *in_data, int in_pitch, float *out_data, int out_pitch, int delta_x, int delta_y, int img_w, int img_h)
{
	const int out_x = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;	// x index to patch.
	const int out_y = IMUL(blockDim.y, blockIdx.y) + threadIdx.y;	// y index to patch.
	const int in_x  = IMUL(delta_x, out_x);
	const int in_y  = IMUL(delta_y, out_y);
	const float *fin = PITCH_GET(in_data, in_pitch, in_y);
	float *fout = PITCH_GET(out_data, out_pitch, out_y);
	fout[out_x] = fin[in_x];
	//fout[out_x] = 1.33333f;
}

// Decimate by factors in x and y.
void svlCudaDecimate(svlCudaPitch &a, svlCudaPitch &b, int delta_x, int delta_y, int img_w, int img_h)
{
	if ( img_w == -1 ) img_w = a.w;
	if ( img_h == -1 ) img_h = b.h;
	dim3 block(16, 16);
	dim3 grid( iDivUp( img_w/delta_x, block.x), iDivUp( img_h/delta_y, block.y) );
	decimate_slice<<< grid, block >>>(a.data, a.pitch, b.data, b.pitch, delta_x, delta_y, img_w, img_h);
}

// Calculate max over an upper-left-centered window.
//template<int NUM_COLS, int NUM_ROWS>
__device__ void max_slice(float *din, int din_pitch, float *res, int res_pitch, int data_w, int data_h, int NUM_COLS, int NUM_ROWS)
{
	// Load din data into shared cache.
	__shared__ float data[ 1024 ];
	const int data_width = 32;
	const int data_size = IMUL(data_width, blockDim.y);
	const int x = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;	// x index to patch.
	const int y = IMUL(blockDim.y, blockIdx.y) + threadIdx.y;	// y index to patch.

	// Read in data.
	const int ix = IMUL(data_width, threadIdx.y) + threadIdx.x;		// Shared thread memory index.
	const float *my_din = PITCH_GET(din, din_pitch, y);				// Calculate my row of din.
	data[ix] = my_din[x];
	// Load rightwards data
	const bool right_on = x + blockDim.x < data_w;
	data[ ix + blockDim.x ] = right_on ? my_din[ x + blockDim.x ] : -FLT_MAX;
	// Load downwards data
	if ( threadIdx.y < NUM_ROWS ) {
		if ( y + blockDim.y < data_h ) {
			// Read in upper entries
			const float *din_down = PITCH_GET(din, din_pitch, y + blockDim.y);
			data[ ix + data_size ] = din_down[x];
			// Read in rightward entry.
			data[ ix + data_size + blockDim.x ] = right_on ? din_down[x + blockDim.x] : -FLT_MAX;
		} else {
			data[ ix + data_size ] = data[ ix + data_size + blockDim.x ] = -FLT_MAX;
		}
	}

	__syncthreads();

	// Compute my max result.
	float gf = -FLT_MAX;
#pragma unroll
	for ( int i=0; i<NUM_ROWS; ++i ) {
#pragma unroll
		for ( int j=0; j<NUM_COLS; ++j ) {
			gf = fmaxf(gf, data[ ix + IMUL(data_width, i) + j ]);
			/*if ( data[ ix + IMUL(data_width, i) + j ] > gf )
				gf = data[ ix + IMUL(data_width, i) + j ];*/
		}
	}

	// Calculate my row of the results.
	float *my_res = PITCH_GET(res, res_pitch, y);
	my_res[x] = gf;
}

// The above without a covering grid (no penalty for inlining a __device__ function).
__global__ void max_NoCG_slice(float *din, int din_pitch, float *res, int res_pitch, int data_w, int data_h, int NUM_COLS, int NUM_ROWS)
{
	max_slice(din, din_pitch, res, res_pitch, data_w, data_h, NUM_COLS, NUM_ROWS);
}
// The above with a covering grid.
__global__ void max_CG_slice(float *din, int din_pitch, float *res, int res_pitch, int data_w, int data_h, int NUM_COLS, int NUM_ROWS)
{
	if ( covering_grid2[ IMUL(blockIdx.y, gridDim.x) + blockIdx.x ] != 0.0f ) {
		max_slice(din, din_pitch, res, res_pitch, data_w, data_h, NUM_COLS, NUM_ROWS);
	}
}

// Calculate max over an upper-left-centered window.
template<int NUM_COLS, int NUM_ROWS>
__device__ void max_template_slice(float *din, int din_pitch, float *res, int res_pitch, int data_w, int data_h)
{
	// Load din data into shared cache.
	__shared__ float data[ 1024 ];
	const int data_width = 32;
	const int data_size = IMUL(data_width, blockDim.y);
	const int x = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;	// x index to patch.
	const int y = IMUL(blockDim.y, blockIdx.y) + threadIdx.y;	// y index to patch.

	// Read in data.
	const int ix = IMUL(data_width, threadIdx.y) + threadIdx.x;		// Shared thread memory index.
	const float *my_din = PITCH_GET(din, din_pitch, y);				// Calculate my row of din.
	data[ix] = my_din[x];
	// Load rightwards data
	const bool right_on = x + blockDim.x < data_w;
	data[ ix + blockDim.x ] = right_on ? my_din[ x + blockDim.x ] : -FLT_MAX;
	// Load downwards data
	if ( threadIdx.y < NUM_ROWS ) {
		if ( y + blockDim.y < data_h ) {
			// Read in upper entries
			const float *din_down = PITCH_GET(din, din_pitch, y + blockDim.y);
			data[ ix + data_size ] = din_down[x];
			// Read in rightward entry.
			data[ ix + data_size + blockDim.x ] = right_on ? din_down[x + blockDim.x] : -FLT_MAX;
		} else {
			data[ ix + data_size ] = data[ ix + data_size + blockDim.x ] = -FLT_MAX;
		}
	}

	__syncthreads();

	// Compute my max result.
	float gf = -FLT_MAX;
#pragma unroll
	for ( int i=0; i<NUM_ROWS; ++i ) {
#pragma unroll
		for ( int j=0; j<NUM_COLS; ++j ) {
			gf = fmaxf(gf, data[ ix + IMUL(data_width, i) + j ]);
			/*if ( data[ ix + IMUL(data_width, i) + j ] > gf )
				gf = data[ ix + IMUL(data_width, i) + j ];*/
		}
	}

	// Calculate my row of the results.
	float *my_res = PITCH_GET(res, res_pitch, y);
	my_res[x] = gf;
}

// The above without a covering grid (no penalty for inlining a __device__ function).
template<int NUM_COLS, int NUM_ROWS>
__global__ void max_template_NoCG_slice(float *din, int din_pitch, float *res, int res_pitch, int data_w, int data_h)
{
	max_template_slice<NUM_COLS, NUM_ROWS>(din, din_pitch, res, res_pitch, data_w, data_h);
}
// The above with a covering grid.
template<int NUM_COLS, int NUM_ROWS>
__global__ void max_template_CG_slice(float *din, int din_pitch, float *res, int res_pitch, int data_w, int data_h)
{
	if ( covering_grid2[ IMUL(blockIdx.y, gridDim.x) + blockIdx.x ] != 0.0f ) {
		max_template_slice<NUM_COLS, NUM_ROWS>(din, din_pitch, res, res_pitch, data_w, data_h);
	}
}

// Perform window-based max.
void svlCudaMax(svlCudaPitch &pitch_in, svlCudaPitch &pitch_out, int max_w, int max_h, int img_w, int img_h, svlCudaCoveringGrid *cudagrid)
{
	dim3 block(16, 16);
	if ( img_w == -1 ) img_w = pitch_in.w;
	if ( img_h == -1 ) img_h = pitch_in.h;
	dim3 grid( iDivUp(img_w, block.x), iDivUp(img_h, block.y) );

	// Verify that grids are of the same size.
	if ( cudagrid == NULL || !sizesMatch(cudagrid, grid) ) {
		if ( max_w == 7 && max_h == 7 ) {
			max_template_NoCG_slice <7,7> <<< grid, block >>>(pitch_in.data, pitch_in.pitch, pitch_out.data, pitch_out.pitch, img_w, img_h);
		}
		else if ( max_w == 6 && max_h == 7 ) {
			max_template_NoCG_slice <6,7> <<< grid, block >>>(pitch_in.data, pitch_in.pitch, pitch_out.data, pitch_out.pitch, img_w, img_h);
		}
		else if ( max_w == 7 && max_h == 6 ) {
			max_template_NoCG_slice <7,6> <<< grid, block >>>(pitch_in.data, pitch_in.pitch, pitch_out.data, pitch_out.pitch, img_w, img_h);
		}
		else if ( max_w == 7 && max_h == 5 ) {
			max_template_NoCG_slice <7,5> <<< grid, block >>>(pitch_in.data, pitch_in.pitch, pitch_out.data, pitch_out.pitch, img_w, img_h);
		}
		else if ( max_w == 5 && max_h == 7 ) {
			max_template_NoCG_slice <5,7> <<< grid, block >>>(pitch_in.data, pitch_in.pitch, pitch_out.data, pitch_out.pitch, img_w, img_h);
		}
		else {
			max_NoCG_slice<<< grid, block >>>(pitch_in.data, pitch_in.pitch, pitch_out.data, pitch_out.pitch, img_w, img_h, max_w, max_h);
		}
	} else {
		// Use covering grid functions.

		// Copy grid to device and use grid'ed version.
		CUDA_SAFE_CALL( cudaMemcpyToSymbol( covering_grid2, cudagrid->data, cudagrid->size()*sizeof(float), 0, cudaMemcpyHostToDevice) );

		//max_CG_slice<<< grid, block >>>(pitch_in.data, pitch_in.pitch, pitch_out.data, pitch_out.pitch, img_w, img_h, max_w, max_h);
		if ( max_w == 7 && max_h == 7 ) {
			max_template_CG_slice <7,7> <<< grid, block >>>(pitch_in.data, pitch_in.pitch, pitch_out.data, pitch_out.pitch, img_w, img_h);
		}
		else if ( max_w == 6 && max_h == 7 ) {
			max_template_CG_slice <6,7> <<< grid, block >>>(pitch_in.data, pitch_in.pitch, pitch_out.data, pitch_out.pitch, img_w, img_h);
		}
		else if ( max_w == 7 && max_h == 6 ) {
			max_template_CG_slice <7,6> <<< grid, block >>>(pitch_in.data, pitch_in.pitch, pitch_out.data, pitch_out.pitch, img_w, img_h);
		}
		else if ( max_w == 7 && max_h == 5 ) {
			max_template_CG_slice <7,5> <<< grid, block >>>(pitch_in.data, pitch_in.pitch, pitch_out.data, pitch_out.pitch, img_w, img_h);
		}
		else if ( max_w == 5 && max_h == 7 ) {
			max_template_CG_slice <5,7> <<< grid, block >>>(pitch_in.data, pitch_in.pitch, pitch_out.data, pitch_out.pitch, img_w, img_h);
		}
		else {
			max_CG_slice<<< grid, block >>>(pitch_in.data, pitch_in.pitch, pitch_out.data, pitch_out.pitch, img_w, img_h, max_w, max_h);
		}
	}
}

// Calculate max in 7-by-7 centered window directly.
__global__ void max_7x7_cent_slice(float *din, int din_pitch, float *res, int res_pitch, int data_w, int data_h)
{
	// Load din data into shared cache.
	__shared__ float data[ 1056 ];
	const int data_width = 48;
	const int data_size = IMUL(data_width, blockDim.y);
	const int x = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;	// x index to patch.
	const int y = IMUL(blockDim.y, blockIdx.y) + threadIdx.y;	// y index to patch.

	// Read in data.
	const int ix = IMUL(data_width, threadIdx.y + 3) + threadIdx.x + 16;		// Shared thread memory index.
	const float *my_din = PITCH_GET(din, din_pitch, y);				// Calculate my row of din.
	data[ix] = my_din[x];
	// Load rightwards data
	const bool right_on = x + blockDim.x < data_w;
	data[ ix + blockDim.x ] = right_on ? my_din[ x + blockDim.x ] : -FLT_MAX;
	// Load leftwards data
	const bool left_on = x >= blockDim.x;
	data[ ix - blockDim.x ] = left_on ? my_din[ x - blockDim.x ] : -FLT_MAX;
	// Load upwards data
	if ( threadIdx.y < 3 ) {
		const int y_off = (threadIdx.y<<1) + 1;
		if ( y - y_off >= 0 ) {
			const float *din_up = PITCH_GET(din, din_pitch, y - y_off);
			data[ ix - IMUL(data_width, y_off) ] = din_up[x];
			// Load sidewards-up data
			data[ ix - IMUL(data_width, y_off) - blockDim.x ] =
				left_on ? din_up[ x - blockDim.x ] : -FLT_MAX;
			data[ ix - IMUL(data_width, y_off) + blockDim.x ] =
				right_on ? din_up[ x + blockDim.x ] : -FLT_MAX;
		} else {
			data[ ix - IMUL(data_width, y_off) ] = -FLT_MAX;
			data[ ix - IMUL(data_width, y_off) - blockDim.x ] = -FLT_MAX;
			data[ ix - IMUL(data_width, y_off) + blockDim.x ] = -FLT_MAX;
		}
	}
	// Load downwards data
	if ( threadIdx.y >= blockDim.y - 3 ) {
		const int y_off = ((blockDim.y - threadIdx.y) <<1) - 1;
		if ( y + y_off < data_h ) {
			const float *din_down = PITCH_GET(din, din_pitch, y + y_off);
			data[ ix + IMUL(data_width, y_off) ] = din_down[x];
			// Load sidewards-up data
			data[ ix + IMUL(data_width, y_off) - blockDim.x ] =
				left_on ? din_down[ x - blockDim.x ] : -FLT_MAX;
			data[ ix + IMUL(data_width, y_off) + blockDim.x ] =
				right_on ? din_down[ x + blockDim.x ] : -FLT_MAX;
		} else {
			data[ ix + IMUL(data_width, y_off) ] = -FLT_MAX;
			data[ ix + IMUL(data_width, y_off) - blockDim.x ] = -FLT_MAX;
			data[ ix + IMUL(data_width, y_off) + blockDim.x ] = -FLT_MAX;
		}
	}

	__syncthreads();

	// Compute my max result.
	float gf = -FLT_MAX;
	//#pragma unroll
	for ( int i=-3; i<=3; ++i ) {
		//#pragma unroll
		for ( int j=-3; j<=3; ++j ) {
			//gf = fmaxf(gf, data[ ix + IMUL(data_width, i) + j ]);
			if ( data[ ix + IMUL(data_width, i) + j ] > gf )
				gf = data[ ix + IMUL(data_width, i) + j ];
		}
	}
	// Calculate my row of the results.
	float *my_res = PITCH_GET(res, res_pitch, y);
	my_res[x] = gf;
}

// Perform window-based normalization.
void svlCudaMax7x7Cent(svlCudaPitch &pitch_in, svlCudaPitch &pitch_out, int img_w, int img_h)
{
	dim3 block(16, 16);
	if ( img_w == -1 ) img_w = pitch_in.w;
	if ( img_h == -1 ) img_h = pitch_in.h;
	dim3 grid( iDivUp(img_w, block.x), iDivUp(img_h, block.y) );
	max_7x7_cent_slice<<< grid, block >>>(pitch_in.data, pitch_in.pitch, pitch_out.data, pitch_out.pitch, img_w, img_h);
}

//// Calculate max in 7-by-7 centered window directly.
//__global__ void max_7x7_slice(float *din, int din_pitch, float *res, int res_pitch, int data_w, int data_h)
//{
//	// Load din data into shared cache.
//	__shared__ float data[ 768 ];
//	const int data_width = 32;
//	const int data_size = IMUL(data_width, blockDim.y);
//	const int x = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;	// x index to patch.
//	const int y = IMUL(blockDim.y, blockIdx.y) + threadIdx.y;	// y index to patch.
//
//	// Read in data.
//	const int ix = IMUL(data_width, threadIdx.y) + threadIdx.x;		// Shared thread memory index.
//	const float *my_din = PITCH_GET(din, din_pitch, y);				// Calculate my row of din.
//	data[ix] = my_din[x];
//	// Load rightwards data
//	const bool right_on = x + blockDim.x < data_w;
//	data[ ix + blockDim.x ] = right_on ? my_din[ x + blockDim.x ] : -FLT_MAX;
//	// Load downwards data
//	if ( threadIdx.y >= blockDim.y - 6 ) {
//		const int y_off = ((blockDim.y - threadIdx.y) <<1) - 1;
//		if ( y + y_off < data_h ) {
//			const float *din_down = PITCH_GET(din, din_pitch, y + y_off);
//			data[ ix + IMUL(data_width, y_off) ] = din_down[x];
//			// Load sidewards-up data
//			data[ ix + IMUL(data_width, y_off) + blockDim.x ] =
//				right_on ? din_down[ x + blockDim.x ] : -FLT_MAX;
//		} else {
//			data[ ix + IMUL(data_width, y_off) ] = -FLT_MAX;
//			data[ ix + IMUL(data_width, y_off) + blockDim.x ] = -FLT_MAX;
//		}
//	}
//
//	__syncthreads();
//
//	// Compute my max result.
//	float gf = -FLT_MAX;
//#pragma unroll
//	for ( int i=0; i<7; ++i ) {
//#pragma unroll
//		for ( int j=0; j<7; ++j ) {
//			//gf = fmaxf(gf, data[ ix + IMUL(data_width, i) + j ]);
//			if ( data[ ix + IMUL(data_width, i) + j ] > gf )
//				gf = data[ ix + IMUL(data_width, i) + j ];
//		}
//	}
//	// Calculate my row of the results.
//	float *my_res = PITCH_GET(res, res_pitch, y);
//	my_res[x] = gf;
//}

// Calculate max in 7-by-7 upper-left-based window directly.
__device__ void max_7x7_slice(float *din, int din_pitch, float *res, int res_pitch, int data_w, int data_h)
{
	// Load din data into shared cache.
	__shared__ float data[ 768 ];
	const int data_width = 32;
	const int data_size = IMUL(data_width, blockDim.y);
	const int x = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;	// x index to patch.
	const int y = IMUL(blockDim.y, blockIdx.y) + threadIdx.y;	// y index to patch.

	// Read in data.
	const int ix = IMUL(data_width, threadIdx.y) + threadIdx.x;		// Shared thread memory index.
	const float *my_din = PITCH_GET(din, din_pitch, y);				// Calculate my row of din.
	data[ix] = my_din[x];
	// Load rightwards data
	const bool right_on = x + blockDim.x < data_w;
	data[ ix + blockDim.x ] = right_on ? my_din[ x + blockDim.x ] : -FLT_MAX;
	// Load downwards data
	if ( threadIdx.y >= blockDim.y - 6 ) {
		const int y_off = ((blockDim.y - threadIdx.y) <<1) - 1;
		if ( y + y_off < data_h ) {
			const float *din_down = PITCH_GET(din, din_pitch, y + y_off);
			data[ ix + IMUL(data_width, y_off) ] = din_down[x];
			// Load sidewards-up data
			data[ ix + IMUL(data_width, y_off) + blockDim.x ] =
				right_on ? din_down[ x + blockDim.x ] : -FLT_MAX;
		} else {
			data[ ix + IMUL(data_width, y_off) ] = -FLT_MAX;
			data[ ix + IMUL(data_width, y_off) + blockDim.x ] = -FLT_MAX;
		}
	}

	__syncthreads();

	// Compute my max result.
	float gf = -FLT_MAX;
#pragma unroll
	for ( int i=0; i<7; ++i ) {
#pragma unroll
		for ( int j=0; j<7; ++j ) {
			//gf = fmaxf(gf, data[ ix + IMUL(data_width, i) + j ]);
			if ( data[ ix + IMUL(data_width, i) + j ] > gf )
				gf = data[ ix + IMUL(data_width, i) + j ];
		}
	}
	// Calculate my row of the results.
	float *my_res = PITCH_GET(res, res_pitch, y);
	my_res[x] = gf;
}

// The above without a covering grid (no penalty for inlining a __device__ function).
__global__ void max_7x7_NoCG_slice(float *din, int din_pitch, float *res, int res_pitch, int data_w, int data_h)
{
	max_7x7_slice(din, din_pitch, res, res_pitch, data_w, data_h);
}
// The above with a covering grid.
__global__ void max_7x7_CG_slice(float *din, int din_pitch, float *res, int res_pitch, int data_w, int data_h)
{
	if ( covering_grid2[ IMUL(blockIdx.y, gridDim.x) + blockIdx.x ] != 0.0f ) {
		max_7x7_slice(din, din_pitch, res, res_pitch, data_w, data_h);
	}
}

// Perform window-based normalization.
void svlCudaMax7x7(svlCudaPitch &pitch_in, svlCudaPitch &pitch_out, int img_w, int img_h, svlCudaCoveringGrid *cudagrid)
{
	dim3 block(16, 16);
	if ( img_w == -1 ) img_w = pitch_in.w;
	if ( img_h == -1 ) img_h = pitch_in.h;
	dim3 grid( iDivUp(img_w, block.x), iDivUp(img_h, block.y) );

	// Verify that grids are of the same size.
	if ( cudagrid != NULL && sizesMatch(cudagrid, grid) ) {
		// Copy grid to device and use grid'ed version.
		CUDA_SAFE_CALL( cudaMemcpyToSymbol( covering_grid2, cudagrid->data, cudagrid->grid_x*cudagrid->grid_y*sizeof(float), 0, cudaMemcpyHostToDevice) );
		max_7x7_CG_slice<<< grid, block >>>(pitch_in.data, pitch_in.pitch, pitch_out.data, pitch_out.pitch, img_w, img_h);
	} else {
		max_7x7_NoCG_slice<<< grid, block >>>(pitch_in.data, pitch_in.pitch, pitch_out.data, pitch_out.pitch, img_w, img_h);
	}
}


// Calculate min and max in 4-by-4 window directly.
// TODO: Templatize and benchmark?
__global__ void minmax_4x4_slice(float *din, int din_pitch, float *min_res, int min_pitch, float *max_res, int max_pitch, int data_w, int data_h)
{
	// Load din data into shared cache.
	__shared__ float data[ 768 ];
	const int data_width = 32;
	const int data_size = IMUL(data_width, blockDim.y);
	const int x = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;	// x index to patch.
	const int y = IMUL(blockDim.y, blockIdx.y) + threadIdx.y;	// y index to patch.

	// Read in data.
	const int ix = IMUL(data_width, threadIdx.y) + threadIdx.x;		// Shared thread memory index.
	const float *my_din = PITCH_GET(din, din_pitch, y);				// Calculate my row of din.
	data[ix] = my_din[x];
	// Load rightwards data
	const bool right_on = x + blockDim.x < data_w;
	data[ ix + blockDim.x ] = right_on ? my_din[ x + blockDim.x ] : my_din[data_w - 1];
	// Load downwards data
	if ( threadIdx.y >= blockDim.y - 3 ) {
		const int y_off = ((blockDim.y - threadIdx.y) <<1) - 1;
		if ( y + y_off < data_h ) {
			const float *din_down = PITCH_GET(din, din_pitch, y + y_off);
			data[ ix + IMUL(data_width, y_off) ] = din_down[x];
			// Load sidewards-up data
			data[ ix + IMUL(data_width, y_off) + blockDim.x ] =
				right_on ? din_down[ x + blockDim.x ] : din_down[data_w - 1];
		} else {
			// Just duplicate values on last valid row.
			const float *din_down = PITCH_GET(din, din_pitch, data_h-1);
			data[ ix + IMUL(data_width, y_off) ] = din_down[x];
			data[ ix + IMUL(data_width, y_off) + blockDim.x ] =
				right_on ? din_down[ x + blockDim.x ] : din_down[data_w - 1];
		}
	}

	__syncthreads();

	// Compute my max result.
	float gf_min = data[ix];
	float gf_max = data[ix];
#pragma unroll
	for ( int i=0; i<4; ++i ) {
#pragma unroll
		for ( int j=0; j<4; ++j ) {
			gf_min = fminf(gf_min, data[ ix + IMUL(data_width, i) + j ]);
			gf_max = fmaxf(gf_max, data[ ix + IMUL(data_width, i) + j ]);
		}
	}
	// Calculate my row of the results.
	float *min_out = PITCH_GET(min_res, min_pitch, y);
	min_out[x] = gf_min;
	float *max_out = PITCH_GET(max_res, max_pitch, y);
	max_out[x] = gf_max;
}

// Calculate min and max over 4-by-4 windows.
void svlCudaMinMax4x4(svlCudaPitch &pitch_in, svlCudaPitch &min_out, svlCudaPitch &max_out, int img_w, int img_h)
{
	dim3 block(16, 16);
	if ( img_w == -1 ) img_w = pitch_in.w;
	if ( img_h == -1 ) img_h = pitch_in.h;
	dim3 grid( iDivUp(img_w, block.x), iDivUp(img_h, block.y) );
	minmax_4x4_slice<<< grid, block >>>(pitch_in.data, pitch_in.pitch, min_out.data, min_out.pitch, max_out.data, max_out.pitch, img_w, img_h);
}


// ****** 32-by-32 min-max functions ******

// Calculate 32-wide min across rows assuming the image has already been min'ed in 4-by-4 windows.
__global__ void min_row32by4_slice(float *min_res, int min_pitch, int data_w, int valid_w)
{
	// Load row into shared memory.
	__shared__ float data[ 512 ];
	const int x = IMUL(valid_w, blockIdx.x) + threadIdx.x;	// x index to patch.
	float *min_out = PITCH_GET(min_res, min_pitch, blockIdx.y);
	data[threadIdx.x] = x < data_w ? min_out[x] : min_out[data_w-1];
	if ( threadIdx.x >= valid_w )
		return; // Early exit for off-of-computation threads.
	
	__syncthreads();

	// Calculate min over 32 spaces by leapfrogging through sub 4-by-4 mins.
	float gf = data[threadIdx.x];
#pragma unroll 7
	for ( int i=1; i<=7; ++i ) {
		gf = fminf(gf, data[ threadIdx.x + (i<<2) ]);
	}
	__syncthreads();

	// Write out result.
	min_out[x] = gf;
	//min_out[x] = -1.0f;
}

// Calculate 32-wide max across rows assuming the image has already been min'ed in 4-by-4 windows.
__global__ void max_row32by4_slice(float *max_res, int max_pitch, int data_w, int valid_w)
{
	// Load row into shared memory.
	__shared__ float data[ 512 ];
	const int x = IMUL(valid_w, blockIdx.x) + threadIdx.x;	// x index to patch.
	float *max_out = PITCH_GET(max_res, max_pitch, blockIdx.y);
	data[threadIdx.x] = x < data_w ? max_out[x] : max_out[data_w-1];
	if ( threadIdx.x >= valid_w )
		return; // Early exit for off-of-computation threads.
	
	__syncthreads();

	// Calculate max over 32 spaces by leapfrogging through sub 4-by-4 maxes.
	float gf = data[threadIdx.x];
#pragma unroll 7
	for ( int i=1; i<=7; ++i ) {
		gf = fmaxf(gf, data[ threadIdx.x + (i<<2) ]);
	}
	__syncthreads();

	// Write out result.
	max_out[x] = gf;
	//max_out[x] = -1.0f;
}

// Calculate 32-wide min down columns assuming the image has already been min'ed in 4-by-4 windows.
// Assumes no runoff in x (i.e., 16 | data_w ).
__global__ void min_col32by4_slice(float *min_res, int min_pitch, int data_h, int valid_h, int y_off)
{
	// Load row into shared memory.
	__shared__ float data[ 512 ];
	const int data_width = 16;
	// Use stuttering order of threads in y.
	const int y = y_off + ((IMUL(valid_h, blockIdx.y) + threadIdx.y)<<2);	// Crazy y get.
	const int x = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;			// Standard x value get.
	// Shared data index
	const int ix = IMUL(data_width, threadIdx.y) + threadIdx.x;
	// Read in data
	float *min_out = y < data_h ? PITCH_GET(min_res, min_pitch, y) : PITCH_GET(min_res, min_pitch, data_h-1);
	data[ix] =  min_out[x];

	if ( threadIdx.y >= valid_h || y >= data_h )
		return; // Early exit for off-of-computation threads.
	
	__syncthreads();

	// Calculate min over 32 spaces by leapfrogging through sub 4-by-4 mins.
	float gf = data[ix];
#pragma unroll 7
	for ( int i=1; i<=7; ++i ) {
		gf = fminf(data[ ix + IMUL(data_width,i) ], gf);
	}
	__syncthreads();

	// Write out result.
	min_out[x] = gf;
}

// Calculate 32-wide max down columns assumaxg the image has already been max'ed in 4-by-4 windows.
// Assumes no runoff in x (i.e., 16 | data_w ).
__global__ void max_col32by4_slice(float *max_res, int max_pitch, int data_h, int valid_h, int y_off)
{
	// Load row into shared memory.
	__shared__ float data[ 512 ];
	const int data_width = 16;
	// Use stuttering order of threads in y.
	const int y = y_off + ((IMUL(valid_h, blockIdx.y) + threadIdx.y)<<2);	// Crazy y get.
	const int x = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;			// Standard x value get.
	// Shared data index
	const int ix = IMUL(data_width, threadIdx.y) + threadIdx.x;
	// Read in data
	float *max_out = y < data_h ? PITCH_GET(max_res, max_pitch, y) : PITCH_GET(max_res, max_pitch, data_h-1);
	data[ix] =  max_out[x];

	if ( threadIdx.y >= valid_h || y >= data_h )
		return; // Early exit for off-of-computation threads.
	
	__syncthreads();

	// Calculate max over 32 spaces by leapfrogging through sub 4-by-4 maxes.
	float gf = data[ix];
#pragma unroll 7
	for ( int i=1; i<=7; ++i ) {
		gf = fmaxf(data[ ix + IMUL(data_width,i) ], gf);
	}
	__syncthreads();

	// Write out result.
	max_out[x] = gf;
}

// Calculate min and max over 32-by-32 windows.
void svlCudaMinMax32x32(svlCudaPitch &pitch_in, svlCudaPitch &min_out, svlCudaPitch &max_out, int img_w, int img_h)
{
	// First calculate internal 4-by-4 min-max windows.
	dim3 block4(16, 16);
	if ( img_w == -1 ) img_w = pitch_in.w;
	if ( img_h == -1 ) img_h = pitch_in.h;
	dim3 grid4( iDivUp(img_w, block4.x), iDivUp(img_h, block4.y) );
	minmax_4x4_slice<<< grid4, block4 >>>(pitch_in.data, pitch_in.pitch, min_out.data, min_out.pitch, max_out.data, max_out.pitch, img_w, img_h);

	// Do row-wise sub operation to get 4-by-32-block min-maxes.
	// TODO: Don't always need to halve image size: only when over thread limit.
	dim3 block_row32( iAlignUp( (img_w>>1) + 28, 16), 1 );
	dim3 grid_row32( 2, img_h );
	min_row32by4_slice<<< grid_row32, block_row32 >>>(min_out.data, min_out.pitch, img_w, img_w>>1);
	max_row32by4_slice<<< grid_row32, block_row32 >>>(max_out.data, max_out.pitch, img_w, img_w>>1);

	// Do col-wise sub operation to get final 32-by-32-block min-maxes.
	dim3 block_col32( 16, 16 );
	int valid_h = block_col32.y - 7;
	dim3 grid_col32( iDivUp( img_w, 16), iDivUp( (img_h>>2), valid_h) );
	/*printf("Block: %d,%d\n", block_col32.x, block_col32.y);
	printf("Grid: %d,%d\n", grid_col32.x, grid_col32.y);*/
	// Since we are using stuttered addresses in y, we need to run this four times for the four offsets.
	min_col32by4_slice<<< grid_col32, block_col32 >>>(min_out.data, min_out.pitch, img_h, valid_h, 0);
	min_col32by4_slice<<< grid_col32, block_col32 >>>(min_out.data, min_out.pitch, img_h, valid_h, 1);
	min_col32by4_slice<<< grid_col32, block_col32 >>>(min_out.data, min_out.pitch, img_h, valid_h, 2);
	min_col32by4_slice<<< grid_col32, block_col32 >>>(min_out.data, min_out.pitch, img_h, valid_h, 3);
	// Do maxes:
	max_col32by4_slice<<< grid_col32, block_col32 >>>(max_out.data, max_out.pitch, img_h, valid_h, 0);
	max_col32by4_slice<<< grid_col32, block_col32 >>>(max_out.data, max_out.pitch, img_h, valid_h, 1);
	max_col32by4_slice<<< grid_col32, block_col32 >>>(max_out.data, max_out.pitch, img_h, valid_h, 2);
	max_col32by4_slice<<< grid_col32, block_col32 >>>(max_out.data, max_out.pitch, img_h, valid_h, 3);
}

// Resize a pitch.
__global__ void resize2D_slice(float *din, int din_pitch, float *dout, int dout_pitch, float sca, int img_w, int img_h)
{
	// Ouput coordinates.
	const int x = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;
	const int y = IMUL(blockDim.y, blockIdx.y) + threadIdx.y;
	if ( x >= img_w || y >= img_h )
		return;

	// Input coordinates.
	float a = float(x) * sca;
	float b = float(y) * sca;
	const int in_x = int(a), in_y = int(b);
	a -= float(in_x);
	b -= float(in_y);
	// in_x and in_y are now the base coordinates of the interpolation window, with a and b the interp factors in x and y.

	const float *din_low = PITCH_GET(din, din_pitch, in_y);
	const float *din_high = PITCH_GET(din, din_pitch, in_y + 1);
	float *doutf = PITCH_GET(dout, dout_pitch, y);

	doutf[x] = (1.0f - a) * (1.0f - b) * din_low[in_x]
			 + a * (1.0f - b) * din_low[in_x + 1]
			 + (1.0f - a) * b * din_high[in_x]
			 + a * b * din_high[in_x + 1];
}
// Used shared data.
// WARNING: Will not work for scales >= 2.0f-, as this will only read in at most a 32x32 window.
__global__ void resize2D_SD_slice(float *din, int din_pitch, float *dout, int dout_pitch, float x_sca, float y_sca, int img_w, int img_h)
{
	__shared__ float data[ 1024 ];

	// Ouput coordinates.
	const int x_base = IMUL(blockDim.x, blockIdx.x);
	const int x = x_base + threadIdx.x;
	const int y_base = IMUL(blockDim.y, blockIdx.y);
	const int y = y_base + threadIdx.y;

	// Input coordinates.
	const int read_x_base = int( float(x_base) * x_sca ),
			  read_y_base = int( float(y_base) * y_sca ); // Base coordinates for window we read in.
	float a = float(x) * x_sca;
	float b = float(y) * y_sca;
	const int in_x = int(a) - read_x_base, in_y = int(b) - read_y_base;
	a -= floorf(a); // Optimal floating-point round-down function.
	b -= floorf(b);
	// in_x and in_y are now the base coordinates of the interpolation window in shared memory, with a and b the interp factors in x and y.

	// Find maximum dimensions of data window that will have to be read in.
	int max_dim = int(fmaxf( ceil(blockDim.x * x_sca) + 1.0f, ceil(blockDim.y * y_sca) + 1.0f ));

	// Read in data.
	const int data_width = 32;
	const float *dinf = PITCH_GET(din, din_pitch, read_y_base + threadIdx.y );
	const int ix = IMUL(threadIdx.y, data_width) + threadIdx.x;
	data[ix] = dinf[ read_x_base + threadIdx.x ];
	// Read in rightwards data.
	const bool right_on = threadIdx.x + blockDim.x < max_dim;
	if ( right_on ) {
		data[ ix + blockDim.x ] = dinf[ read_x_base + blockDim.x + threadIdx.x ];
	}
	// Right in upwards data.
	if ( threadIdx.y + blockDim.y < max_dim ) {
		const float *dinf_up = PITCH_GET(din, din_pitch, read_y_base + blockDim.y + threadIdx.y );
		data[ ix + IMUL(blockDim.y, data_width) ] = dinf_up[ read_x_base + threadIdx.x ];
		// Read in rightwards data.
		if ( right_on ) {
			data[ ix + IMUL(blockDim.y, data_width) + blockDim.x ] = dinf_up[ read_x_base + blockDim.x + threadIdx.x ];
		}
	}

	if ( x > img_w || y > img_h )
		return;

	__syncthreads();

	// Compute and right out result.
	const int in_ix = IMUL(data_width, in_y) + in_x;
	float *doutf = PITCH_GET(dout, dout_pitch, y);
	doutf[x] = (1.0f - a) * (1.0f - b) * data[ in_ix ]
			 + a * (1.0f - b) * data[ in_ix + 1 ]
			 + (1.0f - a) * b * data[ in_ix + data_width ]
			 + a * b * data[ in_ix + data_width + 1 ];

	/*// Debug output
	const int x = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;
	const int y = IMUL(blockDim.y, blockIdx.y) + threadIdx.y;
	if ( x > img_w || y > img_h )
		return;

	const float *dinf = PITCH_GET(din, din_pitch, y);
	float *doutf = PITCH_GET(dout, dout_pitch, y);
	doutf[x] = dinf[x];*/
}

// Resize a 2D pitch.
void svlCudaResize2D(svlCudaPitch &pin, svlCudaPitch &pout, float scale, int img_w, int img_h)
{
	if ( scale >= 1.0f ) {
		printf("svlCudaResize2D only scales down; factor of %0.2f won't work.\n", scale);
		assert(false);
	}
	if ( img_w == -1 ) img_w = pin.w;
	if ( img_h == -1 ) img_h = pin.h;
	// Apply scale.
	int new_img_w = int( float(img_w) * scale );
	int new_img_h = int( float(img_h) * scale );
	float x_scale = float(img_w)/float(new_img_w);
	float y_scale = float(img_h)/float(new_img_h);
	//printf("Rescaling to: %d,%d\n", img_w, img_h);
	dim3 block(16, 16);
	dim3 grid( iDivUp(new_img_w, block.x), iDivUp(new_img_h, block.y) );
	resize2D_SD_slice<<< grid, block >>>(pin.data, pin.pitch, pout.data, pout.pitch, x_scale, y_scale, new_img_w, new_img_h);
}

// Add two pitches in quadrature.
__global__ void add_quad_slice(float *din1, int din1_pitch, float *din2, int din2_pitch, int img_w, int img_h)
{
	const int x = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;
	const int y = IMUL(blockDim.y, blockIdx.y) + threadIdx.y;

	if ( x < img_w && y < img_h ) {
		// Compute result.
		float *f1 = PITCH_GET(din1, din1_pitch, y);
		const float *f2 = PITCH_GET(din2, din2_pitch, y);
		const float ff1 = f1[x] * f1[x];
		const float ff2 = f2[x] * f2[x];
		f1[x] = __powf( ff1 + ff2, 0.5f );
	}
}

// Add two pitches in quadrature.
void svlCudaAddQuadrature(svlCudaPitch *a, svlCudaPitch *b, int img_w, int img_h)
{
	dim3 block(16, 16);
	dim3 grid( iDivUp(img_w, block.x), iDivUp(img_h, block.y) );
	if ( img_w == -1 ) img_w = a->w;
	if ( img_h == -1 ) img_h = a->h;
	add_quad_slice<<< grid, block >>>(a->data, a->pitch, b->data, b->pitch, img_w, img_h);
}


