/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlCudaSlidingWindowDetector.h
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**              Sid Batra <sidbatra@stanford.edu>
**              Ian Goodfellow <ia3n@cs.stanford.edu>
**              Olga Russakovsky <olga@cs.stanford.edu>
** DESCRIPTION:
**  Defines a multi-channel interface for sliding-window boosted-classifier 
**  object detectors. Derived classes must implement feature extraction 
**  functions and may override other functions for more efficient implementation.
**  Uses svlOptions interface for handling training options and basic feature
**  additions. The derived class svlSingleChannelSlidingWindowDetector implements
**  a single channel sliding-window detector.
**
*****************************************************************************/

#pragma once

#include <cassert>
#include <iostream>
#include <vector>
#include <map>

#include "cv.h"
#include "cxcore.h"
#include "ml.h"

#include "svlBase.h"
#include "svlVision.h"
#include "svlCudaCommon.h"
#include "svlCudaConvolution.h"
#include "svlCudaDTree.h"
#include "svlCudaMemHandler.h"
#include "svlCudaMath.h"
#include "../vision/svlSlidingWindowDetector.h"
#include "../vision/svlPatchDefinition.h"

using namespace std;

// svlCudaSlidingWindowDetector ---------------------------------------------------

//#define OLD_SCHOOL

typedef vector<vector<double> > Tsamples;

struct CachedPatch {
	int w, h, N;				// Padded width, height, and original numel.
	float mean, std;			// Patch mean and standard deviation.
	svlCudaLine *patch, *mask;	// Device pointers to padded patch and padded mask.
};

class svlCudaSlidingWindowDetector : public svlSlidingWindowDetector {
protected:
	svlFinalPatchDictionary dict;
	bool use_integral_images;
	int max_channels;
	svlCudaLine *cachedLine;
	vector<CachedPatch> cachedPatches;
	svlCudaDTree *dtree;

	// Pitch identifiers
#ifdef OLD_SCHOOL
	int _IMG, _CONV, _TEMP1, _TEMP2, _INTEG, _INTEG2;
#else
	svlCudaPitch *_CONV, *_TEMP1, *_TEMP2;
	vector<svlCudaPitch*> _IMGS, _INTEGS, _INTEGS2;
#endif

public:
	svlCudaSlidingWindowDetector(int max_img_w, int max_img_h, int max_channels, const char *name = NULL, int w = 32, int h = 32);
	svlCudaSlidingWindowDetector(int max_img_w, int max_img_h, int max_channels, const svlCudaSlidingWindowDetector& c);
	virtual ~svlCudaSlidingWindowDetector();

	// evaluate classifer over all shifts and scales
	svlObject2dFrame cudaClassifyImage(const vector<IplImage *>& images, int outputOffset = 0);
	svlObject2dFrame cudaClassifyImage(const vector<IplImage *>& images,
		const vector<CvPoint>& locations, 
		bool sparse = false);

	// Find interesting windows.
	void createInterestingWindowLocations(svlCudaPitch *img, int img_w, int img_h, vector<CvPoint>& locations) const;
	void createInterestingWindowLocations(IplImage *image, vector<CvPoint>& locations) const;

	// Pad patches and cache GPU-side for only device-to-device copies during computation.
	void padAndCachePatches();
	
	// Load a decision tree.
	void loadCudaDTree(const char *filename);

	inline bool readDictionary(const char *filename) {
		return dict.read(filename);
	}
	inline void printDictionary() {
		dict.print();
	}

	// Nullify previous virtual members.
public:
	virtual svlObject2dFrame classifyImage(const vector<IplImage *>& images,
		svlCudaMemHandler *cudamem = NULL) { return svlObject2dFrame(); }
	virtual svlObject2dFrame classifyImage(const vector<IplImage *>& images,
		const vector<CvPoint>& locations, 
		svlCudaMemHandler *cudamem = NULL,
		bool sparse = false) { return svlObject2dFrame(); }
protected:
	virtual void computePatchFeatureVector(vector<double> &fv, 
		const vector<IplImage*>& images) {}
	virtual void computeImageFeatureVectorsCuda(Tsamples& fv,
		const vector<CvPoint>& locations,
		const vector<IplImage*>& images,
		svlCudaMemHandler *cudamem,
		svlCudaCoveringGrid *cudagrid = NULL,
		bool sparse = false) {}
	virtual void computeImageFeatureVectors(Tsamples& fv,
		const vector<CvPoint>& locations,
		const vector<IplImage*>& images,
		bool sparse = false) {}

};
