/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlAlgorithms.h
** AUTHOR(S):   Paul Baumstarck <pbaumstarck@stanford.edu>
** DESCRIPTION:
**  Class implementing generic algorithms (some in template form) used across
**  the SVL codebase.
**
*****************************************************************************/

#pragma once

#include <vector>
#include <list>

#include "cv.h"
#include "cxcore.h"

//#include "svlBase.h"
//#include "svlDevel.h"
//#include "svlVision.h"
#include "svlSortVectorDI.h"
#include "../vision/svlPoint3d.h"

using namespace std;

class svlAlgorithms {
public:
	// no members

public:
	// templated
	template <class V> int partition(vector<V> &vec, int left, int right, int piv);
	template <class V> int select_ordered(vector<V> &vec, int ord);
	template <class V> int select_ordered(vector<V> &vec, int left, int right, int ord);

	// non-templated
	void computeNNNandRes(vector<svlPoint3d> &P, vector<svlPoint3d> &N, vector<vector<int> > &allNN, svlSortVectorDI *resid);
	void computeNNNandResMS(vector<svlPoint3d> &P, vector<svlPoint3d> &N, vector<vector<int> > &allNN, svlSortVectorDI *resid = NULL, int n_iter = 20);
	void computeNormalMats(CvMat *_X, CvMat *_Y, CvMat *_Z, CvMat *_nX, CvMat *_nY, CvMat *_nZ, int n = 5);
};

template <class V>
int svlAlgorithms::partition(vector<V> &vec, int left, int right, int piv) {
	// Partition based on a pivot.
	// vec: the vector to partition
	// left: inclusive left bound on subsection to partition
	// right: inclusive right bound on subsection to partition
	// piv: index to grab pivot value from, left <= piv <= right
	V pivot_val = vec[piv], temp;
	vec[piv] = vec[right];
	vec[right] = pivot_val; // swap pivot out to right side
	int stix = left; // storage index
	for ( int i=left; i<right; ++i ) {
		if ( vec[i] < pivot_val ) {
			temp = vec[stix];
			vec[stix] = vec[i];
			vec[i] = temp;
			++stix;
		}
	}
	temp = vec[stix];
	vec[stix] = vec[right];
	vec[right] = temp;
	return stix;
}

template <class V>
int svlAlgorithms::select_ordered(vector<V> &vec, int ord) {
	// Select ord'th element from vec by partitioning. Reorders vec.
	return select_ordered<V>(vec,0,(int)vec.size()-1,ord);
}

template <class V>
int svlAlgorithms::select_ordered(vector<V> &vec, int left, int right, int ord) {
	// Select the ord'th element from the vector vec by partitioning. Reorders vec.
	// vec: the input vector
	// left: inclusive left bound to search from in vec
	// right: inclusive right bound to search up to in vec
	// ord: the numbered element to select
	int piv;
	for (;;) {
		piv = partition<V>(vec, left, right, ((right-left)>>1)+left);
		if ( piv == ord )
			return piv;
		else if ( piv > ord )
			right = piv-1;
		else
			left = piv+1;
	}
}

