/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlSufficientStats.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**
*****************************************************************************/

#include <cassert>
#include <iostream>
#include <fstream>
#include <limits>
#include <cmath>

// OpenCV libraries
#include "cv.h"
#include "cxcore.h"

// SVL libraries
#include "svlBase.h"
#include "svlSufficientStats.h"
#include "../vision/svlVisionUtils.h"

using namespace std;
#define assert

// svlSufficientStats class --------------------------------------------------

svlSufficientStats::svlSufficientStats(int n) :
    _n(n), _count(0.0), _sum(NULL), _sum2(NULL)
{
    if (_n > 0) {
        clear(_n);
    }
}

svlSufficientStats::svlSufficientStats(const svlSufficientStats& stats) :
    _n(stats._n), _count(stats._count), _sum(NULL), _sum2(NULL)
{
    if (stats._sum != NULL) {
        _sum = cvCloneMat(stats._sum);
    }

    if (stats._sum2 != NULL) {
        _sum2 = cvCloneMat(stats._sum2);
    }
}

svlSufficientStats::~svlSufficientStats()
{
    if (_sum2 != NULL) cvReleaseMat(&_sum2);
    if (_sum != NULL) cvReleaseMat(&_sum);
}

void svlSufficientStats::clear()
{
    _count = 0.0;
    if (_sum != NULL) cvZero(_sum);
    if (_sum2 != NULL) cvZero(_sum2);
}

void svlSufficientStats::clear(int n)
{
    SVL_ASSERT(n > 0);

    if (_sum2 != NULL) cvReleaseMat(&_sum2);
    if (_sum != NULL) cvReleaseMat(&_sum);

    _n = n;
    _count = 0.0;
    _sum = cvCreateMat(_n, 1, CV_64FC1);
    _sum2 = cvCreateMat(_n, _n, CV_64FC1);
    cvZero(_sum);
    cvZero(_sum2);
}
    
void svlSufficientStats::save(const char *filename) const
{
    SVL_ASSERT(filename != NULL);

    ofstream ofs(filename);
    SVL_ASSERT(!ofs.fail());
    this->save(ofs);
    ofs.close();
}

void svlSufficientStats::save(ostream& os) const
{
    os << _n << "\n\n";
    os << _count << "\n\n";
    writeMatrix(_sum, os);
    os << "\n";
    writeMatrix(_sum2, os);
}

void svlSufficientStats::load(const char *filename)
{
    SVL_ASSERT(filename != NULL);

    ifstream ifs(filename);
    SVL_ASSERT(!ifs.fail());
    this->load(ifs);
    SVL_ASSERT(!ifs.fail());
    ifs.close();
}

void svlSufficientStats::load(istream& is)
{
    int n;
    is >> n;
    this->clear(n);
    
    is >> _count;
    readMatrix(_sum, is);
    readMatrix(_sum2, is);
}

void svlSufficientStats::accumulate(const vector<double>& x, double w)
{
    SVL_ASSERT_MSG((int)x.size() == _n, x.size() << " != " << _n);
    for (int i = 0; i < _n; i++) {
        CV_MAT_ELEM(*_sum, double, i, 0) += w * x[i];
        for (int j = i; j < _n; j++) {
            CV_MAT_ELEM(*_sum2, double, i, j) += w * x[i] * x[j];
            CV_MAT_ELEM(*_sum2, double, j, i) =
                CV_MAT_ELEM(*_sum2, double, i, j);
        }
    }
    _count += w;
}

void svlSufficientStats::accumulate(const vector<vector<double> >& x, double w)
{
    for (vector<vector<double> >::const_iterator it = x.begin();
         it != x.end(); ++it) {
        accumulate(*it, w);
    }
}

void svlSufficientStats::accumulate(const svlSufficientStats& stats, double w)
{
    SVL_ASSERT(stats._n == _n);
    
    _count += w * stats._count;
    for (int i = 0; i < _n; i++) {
        CV_MAT_ELEM(*_sum, double, i, 0) += w * CV_MAT_ELEM(*stats._sum, double, i, 0);
        for (int j = 0; j < _n; j++) {
            CV_MAT_ELEM(*_sum2, double, i, j) += w * 
                CV_MAT_ELEM(*stats._sum2, double, i, j);
        }
    }
}

void svlSufficientStats::subtract(const vector<double>& x, double w)
{
    SVL_ASSERT((int)x.size() == _n);
    SVL_ASSERT(_count >= w);

    for (int i = 0; i < _n; i++) {
        CV_MAT_ELEM(*_sum, double, i, 0) -= w * x[i];
        for (int j = i; j < _n; j++) {
            CV_MAT_ELEM(*_sum2, double, i, j) -= w * x[i] * x[j];
            CV_MAT_ELEM(*_sum2, double, j, i) =
                CV_MAT_ELEM(*_sum2, double, i, j);
        }
    }
    _count -= w;
}

void svlSufficientStats::subtract(const vector<vector<double> >& x, double w)
{
    SVL_ASSERT(w * x.size() <= _count);

    for (vector<vector<double> >::const_iterator it = x.begin();
         it != x.end(); ++it) {
        subtract(*it, w);
    }
}

void svlSufficientStats::subtract(const svlSufficientStats& stats, double w)
{
    SVL_ASSERT(stats._n == _n);
    SVL_ASSERT(w * stats._count <= _count);
    
    _count -= w * stats._count;
    for (int i = 0; i < _n; i++) {
        CV_MAT_ELEM(*_sum, double, i, 0) -= w * CV_MAT_ELEM(*stats._sum, double, i, 0);
        for (int j = 0; j < _n; j++) {
            CV_MAT_ELEM(*_sum2, double, i, j) -= w * 
                CV_MAT_ELEM(*stats._sum2, double, i, j);
        }
    }    
}

// standard operators
svlSufficientStats& svlSufficientStats::operator=(const svlSufficientStats& stats)
{
    // check for self assignment
    if (this->_sum == stats._sum)
        return *this;

    if (_sum2 != NULL) cvReleaseMat(&_sum2);
    if (_sum != NULL) cvReleaseMat(&_sum);

    _n = stats._n;
    _count = stats._count;
    _sum = NULL;
    _sum2 = NULL;

    if (stats._sum != NULL) {
        _sum = cvCloneMat(stats._sum);
    }

    if (stats._sum2 != NULL) {
        _sum2 = cvCloneMat(stats._sum2);
    }

    return *this;
}

// svlCondSufficientStats class ----------------------------------------------

svlCondSufficientStats::svlCondSufficientStats(int n, int k) :
    _n(n), _k(k) {
    if (_k > 0) {
        _stats.resize(_k, svlSufficientStats(_n));
    }
}

svlCondSufficientStats::svlCondSufficientStats(const svlCondSufficientStats& condStats) :
    _n(condStats._n), _k(condStats._k), _stats(condStats._stats)
{
    // do nothing
}

svlCondSufficientStats::~svlCondSufficientStats()
{
    // do nothing
}

void svlCondSufficientStats::clear()
{
    for (int y = 0; y < _k; y++) {
        _stats[y].clear();
    }
}

void svlCondSufficientStats::clear(int n, int k)
{
    SVL_ASSERT((n > 0) && (k > 0));
    
    _n = n;
    _k = k;
    _stats.resize(_k);
    for (int y = 0; y < _k; y++) {
        _stats[y].clear(_n);
    }
}

void svlCondSufficientStats::save(const char *filename) const
{
    SVL_ASSERT(filename != NULL);

    ofstream ofs(filename);
    SVL_ASSERT(!ofs.fail());

    ofs << _n << " " << _k << "\n\n";
    for (int i = 0; i < _k; i++) {
        _stats[i].save(ofs);
    }
    ofs.close();
}

void svlCondSufficientStats::load(const char *filename)
{
    SVL_ASSERT(filename != NULL);

    ifstream ifs(filename);
    SVL_ASSERT(!ifs.fail());

    int n, k;
    ifs >> n >> k;
    clear(n, k);
    
    for (int i = 0; i < k; i++) {
        _stats[i].load(ifs);
        SVL_ASSERT(!ifs.fail());
    }
    ifs.close();
}

// TODO: move inline
void svlCondSufficientStats::accumulate(const vector<double>& x, int y)
{
    SVL_ASSERT((y >= 0) && (y < _k));
    _stats[y].accumulate(x);
}

void svlCondSufficientStats::accumulate(const vector<vector<double> >& x, int y)
{
    SVL_ASSERT((y >= 0) && (y < _k));
    _stats[y].accumulate(x);
}

void svlCondSufficientStats::accumulate(const vector<vector<double> >& x, const vector<int>& y)
{
    SVL_ASSERT(x.size() == y.size());
    for (int i = 0; i < (int)x.size(); i++) {
        accumulate(x[i], y[i]);
    }    
}

void svlCondSufficientStats::accumulate(const svlSufficientStats& stats, int y)
{
    SVL_ASSERT((y >= 0) && (y < _k));
    _stats[y].accumulate(stats);
}

// TODO: move inline
void svlCondSufficientStats::subtract(const vector<double>& x, int y)
{
    SVL_ASSERT((y >= 0) && (y < _k));
    _stats[y].subtract(x);
}

void svlCondSufficientStats::subtract(const vector<vector<double> >& x, int y)
{
    SVL_ASSERT((y >= 0) && (y < _k));
    _stats[y].subtract(x);
}

void svlCondSufficientStats::subtract(const vector<vector<double> >& x, const vector<int>& y)
{
    SVL_ASSERT(x.size() == y.size());
    for (int i = 0; i < (int)x.size(); i++) {
        subtract(x[i], y[i]);
    }    
}

void svlCondSufficientStats::subtract(const svlSufficientStats& stats, int y)
{
    SVL_ASSERT((y >= 0) && (y < _k));
    _stats[y].subtract(stats);
}

void svlCondSufficientStats::redistribute(int y, int k)
{
    SVL_ASSERT((y >= 0) && (y < _k))
    SVL_ASSERT((k >= 0) && (k < _k));

    _stats[k].accumulate(_stats[y]);
    _stats[y].clear();
}
